from fuzzywuzzy import fuzz
import numpy as np
import pandas as pd
import requests


def score_owners(x, search_name):
    # TODO - more complex rules about scoring and providing info on the relevance of owners
    return np.mean([fuzz.partial_ratio(x["owner_name"], search_name)])


class OwnerQuery:
    """Step 1: Takes owner names and generates a parcel num sql query"""

    def __init__(self, owner_names, include_owners_with_same_mailing_address=True):
        # if passed a single owner name, put it in a list
        if type(owner_names) == str:
            owner_names = [owner_names]
        self.owner_names = owner_names

        self.parcel_num_sql_list = []
        self.owner_df = None
        self.include_owners_with_same_mailing_address = (
            include_owners_with_same_mailing_address
        )
        self._set_owner_df_and_parcel_num_sql()
        self.owners_list = self.owner_df["owner_name"].unique()

        # Step 1.5: At this point, users can manually exclude owner names
        # using the owners_list. (TODO)
        # They would modify the parcel_num_sql_lists below
        all_parcel_num_sql = "\nUNION ALL\n".join(self.parcel_num_sql_list)
        self.parcel_num_sql = f"SELECT distinct(parcel_number) from ({all_parcel_num_sql}) as unique_parcels"

    def _set_owner_df_and_parcel_num_sql(self):
        def _sql_string_replace(x):
            # FOR DIRECT OWNER SEARCHING
            # SOLVING THE ISSUE WHERE OWNERS HAVE & IN THE NAME
            return x.replace("&", "'|| chr(38) ||'")

        owner_dfs = []
        for search_name in self.owner_names:
            owner_name = _sql_string_replace(search_name)
            single_owner_df, single_parcel_num_sql = get_complete_owner_list(
                owner_name,
                include_opa_properties_public=True,
                include_rtt_summary=True,
                include_business_licenses=True,
                opa_properties_public_matches_on_mailing_address=self.include_owners_with_same_mailing_address,
            )
            if not single_owner_df.empty:
                single_owner_df["score"] = single_owner_df.apply(
                    lambda x: score_owners(x, search_name), axis=1
                )
                owner_dfs.append(single_owner_df)
                self.parcel_num_sql_list.append(single_parcel_num_sql)
        self.owner_df = (
            pd.concat(owner_dfs)
            .sort_values("score", ascending=False)
            .drop_duplicates("owner_name")
            if owner_dfs
            else pd.DataFrame(columns=["owner_name"])
        )


class OwnerQueryResult:
    """Step 2: Get a timeline of when the owner owned properties"""

    def __init__(self, parcel_num_sql, owners_list):
        self.parcel_num_sql = parcel_num_sql
        self.owners_list = owners_list

        deeds_df = get_deeds_list(parcel_num_sql)

        def _deed_timeline(parcel_df):
            parcel_df["start_dt"] = parcel_df["deed_date"]
            parcel_df["end_dt"] = parcel_df["start_dt"].shift(-1)
            first_owner = parcel_df.iloc[0].copy(deep=True).to_dict()
            first_owner["likely_owner"] = (
                first_owner["grantors"] if first_owner["grantors"] else "?"
            )
            # if year built is 0000, assume it was constructed at the first deed
            first_owner["start_dt"] = (
                first_owner["year_built"]
                if first_owner["year_built"] != "0000"
                else first_owner["deed_date"]
            )
            first_owner["end_dt"] = first_owner["deed_date"]

            combined_df = pd.concat(
                [
                    parcel_df,
                    pd.DataFrame([first_owner])[
                        ["parcel_number", "likely_owner", "start_dt", "end_dt"]
                    ],
                ],
                sort=False,
            ).sort_values(["parcel_number", "end_dt"])
            combined_df["street_address"] = combined_df["street_address"].bfill()
            combined_df["location"] = combined_df["location"].bfill()
            combined_df["lat"] = combined_df["lat"].bfill()
            combined_df["lng"] = combined_df["lng"].bfill()
            combined_df["unit"] = combined_df["unit"].bfill()
            return combined_df

        timeline_df = deeds_df.groupby("parcel_number").apply(_deed_timeline)[
            [
                "parcel_number",
                "grantees",
                "owner_1",
                "owner_2",
                "likely_owner",
                "start_dt",
                "end_dt",
                "street_address",
                "location",
                "lat",
                "lng",
                "unit",
                "unit_num",
            ]
        ]

        def _check_owner_names_are_equivalent(owner_1, owner_2):
            # owner in opa_properties_public get cut off at 25 chars
            return owner_1 == owner_2 or (
                len(owner_1) == 25 and owner_2.startswith(owner_1)
            )

        def _check_owner_in_list(x, owners_list):
            return any(
                [
                    _check_owner_names_are_equivalent(poss_owner, part_owner)
                    for poss_owner in owners_list
                    for part_owner in x.split(";")
                ]
            )

        owners_timeline_df = timeline_df[
            timeline_df["likely_owner"].apply(
                lambda x: _check_owner_in_list(x, self.owners_list)
            )
        ].sort_index()

        self.timeline_df = timeline_df
        self.owners_timeline_df = owners_timeline_df
        self.deeds_df = deeds_df

    def get_filtered_df(self, df_table, dt_col):
        if df_table.empty:
            return pd.DataFrame()
        else:
            results_df = (
                self.owners_timeline_df.set_index("parcel_number")[
                    ["likely_owner", "start_dt", "end_dt"]
                ]
                .merge(df_table, how="left", on="parcel_number")
                .set_index("parcel_number")
                .dropna(subset=[dt_col])
            )

            results_df[dt_col] = pd.to_datetime(
                results_df[dt_col], utc=True
            ) + pd.Timedelta(
                days=1
            )  # 888035586 off by a day licenses
            results_df["start_dt"] = pd.to_datetime(results_df["start_dt"], utc=True)
            results_df["end_dt"] = pd.to_datetime(results_df["end_dt"], utc=True)
            filtered_results_df = results_df[
                (results_df[dt_col] > results_df["start_dt"])
                & (
                    (
                        pd.isnull(results_df["end_dt"])
                        | (results_df[dt_col] < results_df["end_dt"])
                    )
                )
            ]
            return filtered_results_df


def get_complete_owner_list(
    owner_name,
    include_opa_properties_public=True,
    include_rtt_summary=True,
    include_business_licenses=True,
    opa_properties_public_matches_on_mailing_address=False,
):
    owners_list = []
    parcel_num_subquery_list = []
    # opa_owners
    if include_opa_properties_public:
        opa_properties_public_query = f"""
            SELECT owner_1, owner_2, parcel_number
            FROM opa_properties_public
            WHERE owner_1 like '{owner_name}' or owner_2 like '{owner_name}'
        """
        opa_owners_result = requests.get(
            f"""
            https://phl.carto.com/api/v2/sql?q={opa_properties_public_query}
            """
        ).json()["rows"]
        opa_owner_1_list = list(
            set([r["owner_1"] for r in opa_owners_result if r["owner_1"]])
        )
        opa_owner_2_list = list(
            set([r["owner_2"] for r in opa_owners_result if r["owner_2"]])
        )
        owners_list.extend(
            [
                {"owner_name": y.strip(), "source": "opa_properties_public.owner_1"}
                for x in opa_owner_1_list
                for y in x.split(";")
            ]
            + [
                {"owner_name": y.strip(), "source": "opa_properties_public.owner_2"}
                for x in opa_owner_2_list
                for y in x.split(";")
            ]
        )
        parcel_num_subquery_list.append(
            f"SELECT distinct(inner_search_opa.parcel_number) from ({opa_properties_public_query}) inner_search_opa"
        )
    if opa_properties_public_matches_on_mailing_address:
        opa_properties_public_query = f"""
            SELECT owner_1, owner_2, parcel_number
            FROM opa_properties_public
            WHERE mailing_street in (
                SELECT mailing_street from opa_properties_public
                WHERE owner_1 like '{owner_name}' or owner_2 like '{owner_name}'
            )
            """
        opa_owners_result = requests.get(
            f"""
            https://phl.carto.com/api/v2/sql?q={opa_properties_public_query}
            """
        ).json()["rows"]
        opa_owner_1_list = list(
            set([r["owner_1"] for r in opa_owners_result if r["owner_1"]])
        )
        opa_owner_2_list = list(
            set([r["owner_2"] for r in opa_owners_result if r["owner_2"]])
        )
        owners_list.extend(
            [
                {"owner_name": y.strip(), "source": "opa_properties_public.owner_1"}
                for x in opa_owner_1_list
                for y in x.split(";")
            ]
            + [
                {"owner_name": y.strip(), "source": "opa_properties_public.owner_2"}
                for x in opa_owner_2_list
                for y in x.split(";")
            ]
        )
        parcel_num_subquery_list.append(
            f"SELECT distinct(inner_search_opa.parcel_number) from ({opa_properties_public_query}) inner_search_opa"
        )

    # rtt summary
    if include_rtt_summary:
        rtt_summary_query = f"""
            SELECT grantees, opa_account_num
            FROM rtt_summary
            WHERE grantees like '{owner_name}'
        """
        rtt_summary_result = requests.get(
            f"""
            https://phl.carto.com/api/v2/sql?q={rtt_summary_query}
            """
        ).json()["rows"]
        grantees_list = list(set([r["grantees"] for r in rtt_summary_result]))
        grantors_list = list(set([r["grantees"] for r in rtt_summary_result]))
        owners_list.extend(
            [
                {"owner_name": y.strip(), "source": "rtt_summary.grantees"}
                for x in grantees_list
                for y in x.split(";")
            ]
            + [
                {"owner_name": y.strip(), "source": "rtt_summary.grantors"}
                for x in grantors_list
                for y in x.split(";")
            ]
        )
        parcel_num_subquery_list.append(
            f"SELECT distinct(inner_search_rtt.opa_account_num) from ({rtt_summary_query}) inner_search_rtt"
        )

    # licenses
    if include_business_licenses:
        business_licenses_query = f"""
            SELECT opa_owner, business_name, legalname, licensestatus, opa_account_num
            FROM business_licenses
            WHERE (
                legalname like '{owner_name}' 
                OR business_name like '{owner_name}' 
                /*OR opa_owner like '{owner_name}' opa_owner is the current owner, not the one at the time of the license*/
            ) and licensetype = 'Rental'
        """
        li_owner_result = requests.get(
            f"""
            https://phl.carto.com/api/v2/sql?q={business_licenses_query}
            """
        ).json()["rows"]

        # include opa_owner field if license status is active, which implies that the license occurred
        # while the current owner owned that building
        li_opa_owner_list = list(
            set(
                [
                    r["opa_owner"]
                    for r in li_owner_result
                    if r["opa_owner"] and r["licensestatus"] == "Active"
                ]
            )
        )
        li_legal_name_list = list(
            set([r["legalname"] for r in li_owner_result if r["legalname"]])
        )
        li_business_name_list = list(
            set([r["business_name"] for r in li_owner_result if r["business_name"]])
        )
        owners_list.extend(
            [
                {"owner_name": y.strip(), "source": "licenses.opa_owner"}
                for x in li_opa_owner_list
                for y in x.split(";")
            ]
            + [
                {"owner_name": y.strip(), "source": "licenses.legalname"}
                for x in li_legal_name_list
                for y in x.split(";")
            ]
            + [
                {"owner_name": y.strip(), "source": "licenses.business_name"}
                for x in li_business_name_list
                for y in x.split(";")
            ]
        )
        parcel_num_subquery_list.append(
            f"SELECT distinct(inner_search_li.opa_account_num) from ({business_licenses_query}) inner_search_li"
        )
    if owners_list:
        owner_df = (
            pd.DataFrame(owners_list)
            .drop_duplicates("owner_name")
            .sort_values("owner_name")
        )
        parcel_num_subquery = "\nUNION ALL\n".join(parcel_num_subquery_list)
        parcel_num_sql = f"""
            SELECT distinct(parcel_number)
            FROM opa_properties_public
            where parcel_number in  ({parcel_num_subquery})
        """
        return owner_df, parcel_num_sql
    else:
        return pd.DataFrame(), "SELECT distinct(parcel_number) WHERE 1=0"


def get_deeds_list(parcel_numbers_str):
    """
    Single SQL query attempt

    select estimated_owner, date, parcel_number from

    property assessment table joined to rtt table fields that are used in the where clause

    group by every property

    and get the
    case 1:
        latest grantee before XXX-date
    case2:
        earliest grantor
    case 3:
        owner_1;owner_2
    """
    where = """
    (

        (
            /* The adddress_low between rtt and opa match (usually the case on exact address match) */
            cast(rtt.ADDRESS_LOW as int) >= opa.address_low AND cast(rtt.ADDRESS_LOW as int) <= opa.address_low
        ) OR (
            /* If the opa address is exact but the rtt address is a span like 3608-14 X St */
            cast(rtt.ADDRESS_LOW as int)>=  opa.address_low - (MOD(opa.address_low, 100))
            AND cast(rtt.ADDRESS_HIGH as int) >= MOD(opa.address_low, 100) 
            AND cast(rtt.ADDRESS_LOW as int) <= ( 
                CASE 
                    WHEN opa.address_high is not null
                    THEN opa.address_high + opa.address_low - (MOD(opa.address_low, 100))
                ELSE opa.address_low
                END   
            )
        )
    )

    /* street names match, opa.street_designation sometimes has trailing spaces so use RTRIM */
    /* 182108500 is one of 126000 that has 0Xth in opa and Xth in rtt so need to LTRIM */
    AND (rtt.STREET_NAME = LTRIM(opa.street_name, '0') )
    AND  MOD(cast(rtt.ADDRESS_LOW as int),2) = MOD(opa.address_low,2) /* Check on correct side of street */
    /* 52338901 example of street suffix being wrong so also match on full street address */
    AND (
            rtt.STREET_SUFFIX = RTRIM(opa.street_designation) 
            OR rtt.STREET_ADDRESS = concat(opa.location) 
    )
    AND (
        CASE
            WHEN opa.street_direction is not null THEN (
                rtt.STREET_PREDIR = opa.street_direction or rtt.STREET_POSTDIR = opa.street_direction
            ) 
            ELSE True
        END
    )
    AND (
        CASE
            WHEN opa.suffix is not null and opa.suffix = '2' THEN (
                rtt.ADDRESS_LOW_SUFFIX = '2' or rtt.ADDRESS_LOW_FRAC = '1/2'
            ) 
            /* had to add 'like' becuase 4604R WHITAKER AVE had RR as the rtt.address_low_suffix */
            WHEN opa.suffix is not null THEN  rtt.ADDRESS_LOW_SUFFIX like concat(opa.suffix, '%')
            ELSE rtt.ADDRESS_LOW_SUFFIX is null and rtt.ADDRESS_LOW_FRAC is null
        END
    )
    AND CASE WHEN opa.unit is not null THEN rtt.UNIT_NUM = opa.unit ELSE True END
    AND
    (
        DOCUMENT_TYPE ='DEED' or 
        DOCUMENT_TYPE='DEED SHERIFF' or 
        DOCUMENT_TYPE='DEED OF CONDEMNATION' or 
        DOCUMENT_TYPE='DEED LAND BANK'
    )
    """
    query = f"""
    SELECT opa.owner_1,opa.owner_2, opa.parcel_number, opa.year_built, 
    ST_Y(opa.the_geom) AS lat, ST_X(opa.the_geom) AS lng,
    rtt.opa_account_num, rtt.unit_num,
    rtt.grantors, rtt.grantees, 
    rtt.street_address, rtt.recording_date,
    opa.location, opa.unit,
    CASE
        WHEN rtt.recording_date is null THEN opa.recording_date
        ELSE rtt.recording_date
    END as deed_date,
    CASE
        WHEN rtt.recording_date is null THEN RTRIM(concat(opa.owner_1,';',opa.owner_2), ';')
        ELSE rtt.grantees
    END as likely_owner
    FROM (
        SELECT * FROM rtt_summary
    ) rtt
    RIGHT JOIN (
        SELECT
        the_geom,
        parcel_number,
        year_built,
        owner_1, owner_2,
        location,
        category_code_description, zoning,
        cast(house_number as int) as address_low,
        cast(house_extension as int) as address_high,
        street_direction,
        street_name,
        street_designation,
        unit, 
        suffix, 
        recording_date
        FROM opa_properties_public
    ) opa ON
    {where}
    WHERE opa.parcel_number in ({parcel_numbers_str})
    GROUP BY opa.owner_1,opa.owner_2, opa.parcel_number, opa.year_built, 
    lat, lng,
    rtt.opa_account_num,
        rtt.grantors, rtt.grantees, rtt.street_address, opa.location, opa.unit, rtt.unit_num,
        rtt.recording_date, opa.recording_date
    ORDER BY opa.parcel_number, rtt.recording_date
    """
    params = {"q": query}
    opa_properties_deeds = requests.get(
        f"""https://phl.carto.com/api/v2/sql""", params=params
    ).json()
    try:
        return pd.DataFrame(opa_properties_deeds["rows"])
    except:
        raise ValueError(f"{opa_properties_deeds}\n\n{parcel_numbers_str}\n{query}")



from abc import ABC, abstractmethod
from datetime import datetime
import logging
import pandas as pd
import requests
import time
import sys
import urllib

from .labelled_fields import label_city_owned_properties, CITY_OWNED_PROPERTY_FIELDS
from .additional_links import (
    get_street_view_link,
    get_atlas_link,
    get_property_phila_gov_link,
    get_license_inspections_link,
)
from .search_queries import search_method_sql


logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger(__name__)


def _format_opa_account_numbers_sql(opa_account_numbers):
    if type(opa_account_numbers) == list:
        return ", ".join([f"'{num}'" for num in opa_account_numbers])
    else:
        return opa_account_numbers


class PhillyArcgisQuery(ABC):
    """Query the arcgis server that contains raw open philly data

    Notes
    -----
    - It seems to only return a max of 2000 entries per query.
    """

    def __init__(self, where_sql, table):
        self.where_sql = where_sql
        self.table = table

    def execute(self):
        # out_fields = ",+".join(self.columns + ['opa_account_num']).upper()
        # going to return all out fields for now
        out_fields = "*"
        encoded_sql = urllib.parse.quote_plus(self.where_sql)
        request_url = (
            "https://services.arcgis.com/fLeGjb7u4uXqeF9q/ArcGIS/rest/services/"
            f"{self.table}/FeatureServer/0/query?returnDistinctValues=true"
            f"&returnGeometry=false&f=json&sqlFormat=standard&where={self.where_sql}"
            f"&outFields={out_fields}"
        )
        query_result = requests.get(request_url)
        return PhillyArcgisQueryResult(
            sql=self.where_sql,
            table=self.table,
            query_result=query_result,
            remove_all_city_owned_properties=False,
            validation_field_to_check="features",
        )


class PhillyCartoQuery(ABC):
    """Query the Carto database (modeled after ibis)"""

    def __init__(self, sql):
        self.sql = sql

    def execute(self, remove_all_city_owned_properties=False):
        payload = {"q": self.sql}
        query_result = requests.get("https://phl.carto.com/api/v2/sql", params=payload)
        return PhillyCartoQueryResult(
            sql=self.sql,
            query_result=query_result,
            remove_all_city_owned_properties=remove_all_city_owned_properties,
        )


class PhillyQueryResult(ABC):
    def __init__(
        self,
        query_result,
        remove_all_city_owned_properties=True,
        validation_field_to_check="rows",
        **kwargs,
    ):
        self.kwargs = kwargs
        self.sql = kwargs.get("sql")
        self.query_result = query_result
        self.results_json = self.query_result.json()

        self.remove_all_city_owned_properties = remove_all_city_owned_properties

        if validation_field_to_check not in self.results_json:
            raise ValueError(self.sql, self.results_json)

    def to_json(self):
        return self.results_json

    def _get_dataframe(self, *args, **kwargs):
        raise NotImplementedError

    def to_dataframe(self, dt_column=None):
        df = self._get_dataframe()
        columns = df.columns.values.tolist()
        df = df if df.empty else df.sort_values(columns, ascending=False)

        # adds a year column based on the specified datetime column
        if dt_column:
            df.insert(0, "dt_year", pd.to_datetime(df[dt_column]).dt.year)

        # adds hyperlinks to city websites
        if "location" in columns:
            df["link_cyclomedia_street_view"] = df["location"].apply(
                get_street_view_link
            )
            df["link_license_inspections"] = df["location"].apply(
                get_license_inspections_link
            )
        for opa_account_col in ["opa_account_num", "parcel_number"]:
            if opa_account_col in columns:
                df["link_property_phila_gov"] = df[opa_account_col].apply(
                    get_property_phila_gov_link
                )
                df["link_atlas"] = df[opa_account_col].apply(get_atlas_link)
                break

        """
        # No longer removing city-owned properties
        df = label_city_owned_properties(
            df, remove=self.remove_all_city_owned_properties
        )
        """
        return df


class PhillyCartoQueryResult(PhillyQueryResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_dataframe(self):
        columns = list(self.results_json["fields"].keys())
        return pd.DataFrame(self.results_json["rows"], columns=columns)


class PhillyArcgisQueryResult(PhillyQueryResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_dataframe(self):
        df = pd.DataFrame([r["attributes"] for r in self.results_json["features"]])
        date_cols = [
            r["name"]
            for r in self.results_json["fields"]
            if r["type"] == "esriFieldTypeDate"
        ]
        for col in date_cols:
            df[col] = df[col].apply(
                lambda x: datetime.fromtimestamp(x / 1000.0) if pd.notnull(x) else x
            )
        return df


class PhillyCartoTable(ABC):
    def __init__(
        self,
        cartodb_table_name,
        title=None,
        sql_alias=None,
        open_data_philly_table_url_name=None,
        schema_application_id=None,
        schema_representation_id=None,
    ):
        self.cartodb_table_name = cartodb_table_name
        self.arcgis_table_name = cartodb_table_name.upper()
        self.title = title if title else self.name.title()
        self.sql_alias = sql_alias if sql_alias else self.cartodb_table_name[:3]
        self.default_columns = []  # set by subclass
        self.dt_column = None
        self.open_data_philly_table_url_name = open_data_philly_table_url_name
        self.schema_application_id = schema_application_id
        self.schema_representation_id = schema_representation_id

        self.data_links = self._get_data_links()

        self.default_opa_properties_public_joined_columns = [
            "location",
            "unit",
            "owner_1",
            "owner_2",
            "mailing_care_of",
            "mailing_street",
            "mailing_address_1",
            "mailing_address_2",
            "mailing_city_state",
            "parcel_number",
            "year_built",
        ]
        self.city_owned_prop_cols = list(CITY_OWNED_PROPERTY_FIELDS.keys())

    def __str__(self):
        return f"{self.title}, {self.cartodb_table_name}"

    def list(
        self,
        columns=None,
        where_sql=None,
        order_by_columns=None,
        limit=None,
        offset=None,
    ):
        select_sql = ",".join(columns) if columns else "*"
        where_sql = f"WHERE {where_sql}" if where_sql else ""
        offset_sql = f"OFFSET {offset}" if offset else ""
        limit_sql = f"LIMIT {limit}" if limit else ""
        order_by_sql = (
            "ORDER BY " + ",".join(order_by_columns) if order_by_columns else ""
        )
        return PhillyCartoQuery(
            f"""
            SELECT {select_sql}
            FROM {self.cartodb_table_name} {self.sql_alias}
            {where_sql}
            {limit_sql}
            {offset_sql}
            {order_by_sql}
            """
        ).execute(remove_all_city_owned_properties=False)

    def query_arcgis(
        self,
        where_sql,
        columns=None,
    ):
        return PhillyArcgisQuery(
            where_sql=where_sql,
            table=self.arcgis_table_name,
        ).execute()

    def query_by_opa_account_numbers(
        self,
        opa_account_numbers,
        dt_column=None,
        columns=None,
        opa_properties_public_joined_columns=None,
        remove_all_city_owned_properties=True,
    ):
        """
        Parameters
        ----------
        opa_account_numbers: list of str
            Either a SQL query that returns a list of account numbers,
            or a hardcoed list of account numbersto query for in this table
        dt_column: str
            SQL column to use to extract a year column from. If None uses the default.
        columns: Can either be a list of columns, or the string 'all' which returns
            all available columns. If nothing is passed, it will use the hardcoded
            default columns.
        opa_properties_public_joined_columns
            Columns to include from the properties table that all tables (other than
            the properties table itself) are joined to.

        """
        dt_column = dt_column if dt_column else self.dt_column
        columns = columns if columns else self.default_columns
        col_str = self._get_column_sql_from_param(columns)

        opa_properties_public_joined_columns = (
            opa_properties_public_joined_columns
            if opa_properties_public_joined_columns
            else self.default_opa_properties_public_joined_columns
        )
        joined_col_str = self._get_column_sql_from_param(
            opa_properties_public_joined_columns, sql_alias="opa"
        )

        opa_account_numbers = _format_opa_account_numbers_sql(opa_account_numbers)
        sql = self._get_sql_for_query_by_opa_account_numbers(
            opa_account_numbers=opa_account_numbers,
            col_str=col_str,
            joined_col_str=joined_col_str,
        )
        return PhillyCartoQuery(sql).execute(
            remove_all_city_owned_properties=remove_all_city_owned_properties
        )

    def query_by_single_str_column(
        self,
        search_column,
        search_to_match,
        search_method="starts with",
        result_columns=None,
        limit=None,
    ):
        """
        A more general way to get results by a single column string match.
        This is helpful for autocomplete functionality.

        Parameters
        ----------
        search_column: str
            Column to use for querying
        search_to_match:
            Value to use for querying to match search_column
        search_method: str
            One of: ['contains', 'starts with', 'ends with', 'equals']
        result_columns:
            Other columns to output when executing this query
        """
        result_columns = result_columns if result_columns else self.default_columns
        col_str = self._get_column_sql_from_param(result_columns)
        joined_col_str = self._get_column_sql_from_param(
            self.city_owned_prop_cols, sql_alias="opa"
        )
        search_to_match = search_method_sql(search_to_match, search_method)
        limit_str = f"LIMIT {limit}" if limit else ""
        sql = self._get_sql_for_query_by_single_str_column(
            search_column=search_column,
            col_str=col_str,
            joined_col_str=joined_col_str,
            search_to_match=search_to_match,
            limit_str=limit_str,
        )
        return PhillyCartoQuery(sql).execute()

    def _get_sql_for_query_by_single_str_column(
        self,
        search_column,
        col_str,
        joined_col_str,
        search_to_match,
        limit_str,
    ):
        return f"""
        SELECT {col_str}, {joined_col_str}
        FROM {self.cartodb_table_name} {self.sql_alias}
        LEFT JOIN opa_properties_public opa 
            ON {self.sql_alias}.opa_account_num=opa.parcel_number
        WHERE {self.sql_alias}.{search_column} LIKE '{search_to_match}' 
        {limit_str}
        """

    def _get_sql_for_query_by_opa_account_numbers(
        self, opa_account_numbers, col_str, joined_col_str
    ):
        return f"""
        SELECT {col_str}, {joined_col_str}
        FROM {self.cartodb_table_name} {self.sql_alias}
        LEFT JOIN opa_properties_public opa 
            ON {self.sql_alias}.opa_account_num=opa.parcel_number
        WHERE opa.parcel_number in ({opa_account_numbers})
        """

    def _get_column_sql_from_param(self, columns, sql_alias=None):
        sql_alias = sql_alias if sql_alias else self.sql_alias
        if type(columns) == str and columns == "all":
            return f"{self.sql_alias}.*"
        elif type(columns) == list:
            return ", ".join([f"{sql_alias}.{c}" for c in columns])

    def _get_data_links(self):
        data_links = []
        odb_link = self._get_odb_link()
        if odb_link:
            data_links.append(odb_link)
        cartodb_link = self._get_cartodb_link()
        if cartodb_link:
            data_links.append(cartodb_link)
        schema_link = self._get_schema_link()
        if schema_link:
            data_links.append(schema_link)
        return data_links

    def _get_odb_link(self):
        return (
            (
                "https://www.opendataphilly.org/dataset/"
                + self.open_data_philly_table_url_name
            )
            if self.open_data_philly_table_url_name
            else None
        )

    def _get_schema_link(self, as_json=False):
        if self.schema_representation_id and self.schema_application_id:
            if as_json:
                return (
                    "https://us-api.knack.com/v1/scenes/scene_142/views/view_287/records/"
                    f"export/applications/550c60d00711ffe12e9efc64?type=json"
                    f"&representationdetails_id={self.schema_representation_id}"
                )
            else:
                return (
                    "https://schema.phila.gov/#home/datasetdetails/"
                    f"{self.schema_application_id}/representationdetails/"
                    f"{self.schema_representation_id}/"
                )

    def _get_cartodb_link(self):
        return (
            "https://cityofphiladelphia.github.io/carto-api-explorer/#"
            + self.cartodb_table_name
        )

    def get_schema(self):
        schema_link = self._get_schema_link(as_json=True)
        if schema_link:
            response = requests.get(schema_link)
            return [
                {
                    "column": r["field_17"],
                    "column_for_display": r["field_188"],
                    "description_str": r["field_20_raw"],
                    "description_html": r["field_20"],
                }
                for r in response.json()["records"]
            ]


class RealEstateTaxRevenue(ABC):
    def __init__(self, rate_limit_wait_secs=5):
        self.data_links = [
            "https://www.phila.gov/revenue/realestatetax/",
            "https://github.com/CityOfPhiladelphia/tips-api",
        ]
        self.rate_limit_wait_secs = rate_limit_wait_secs

    def query_by_opa_account_numbers(
        self,
        opa_account_numbers,
        remove_all_city_owned_properties=True,
    ):
        """
        Parameters
        ----------
        """
        if type(opa_account_numbers) == list:
            opa_account_num_list = opa_account_numbers
        else:
            opa_account_numbers_sql = _format_opa_account_numbers_sql(
                opa_account_numbers
            )
            opa_account_num_list = [
                r["parcel_number"]
                for r in PhillyCartoQuery(opa_account_numbers_sql)
                .execute()
                .to_json()["rows"]
            ]
        output_dfs = []
        for account_num in opa_account_num_list:
            url = f"https://api.phila.gov/tips/account/{account_num}"
            response = requests.get(url).json()
            logger.info(
                f"Waiting {self.rate_limit_wait_secs} seconds to not violate rate limit."
            )
            time.sleep(self.rate_limit_wait_secs)
            if "data" in response:
                data = response["data"]
                property_data = data["property"]
                years_data = data["years"]
                df = pd.DataFrame(years_data)
                df["last_payment_posted_date"] = data["lastPaymentPostedDate"]
                df["account_num"] = data["accountNum"]
                for key, val in property_data.items():
                    df[key] = val
                output_dfs.append(df)
        return pd.concat(output_dfs)



from abc import ABC
from .exceptions import (
    SearchMethodNotImplementedError,
    SearchTypeNotImplementedError,
)


def search_method_sql(search_to_match, search_method):
    if search_method == "contains":
        return f"%{search_to_match}%"
    elif search_method == "starts with":
        return f"{search_to_match}%"
    elif search_method == "ends with":
        return f"%{search_to_match}"
    elif search_method == "equals":
        return search_to_match
    else:
        raise SearchMethodNotImplementedError(search_method)


def construct_search_query(
    search_to_match,
    search_type,
    search_method="contains",
):
    """
    Takes in a search query and method of searching, and constructs a SQL subquery
    that when executed provides a list of opa account numbers.
    search_method=['contains', 'starts with', 'ends with', 'equals'],

    Parameters
    ----------
    search_to_match: str
    search_type: str
        One of: ['mailing_address', 'owner', 'location_by_owner',
            'location_by_mailing_address', 'parcel_number'
            ]
    search_method: str
        One of: ['contains', 'starts with', 'ends with', 'equals']
    """
    search_to_match = search_method_sql(search_to_match.upper(), search_method)
    if search_type == "parcel_number":
        opa_account_numbers_sql = f"""
            SELECT opp.parcel_number FROM opa_properties_public opp 
                WHERE (
                    opp.parcel_number LIKE '{search_to_match}'
                )
        """
    elif search_type == "owner":
        opa_account_numbers_sql = f"""
            SELECT opp.parcel_number FROM opa_properties_public opp 
                WHERE (
                    owner_1 LIKE '{search_to_match}' OR owner_2 LIKE '{search_to_match}'
                )
            UNION ALL
            SELECT opa_account_num FROM business_licenses bl, opa_properties_public opp
                WHERE bl.opa_account_num = opp.parcel_number
                AND licensetype = 'Rental'
                AND (
                    legalname LIKE '{search_to_match}' OR business_name LIKE '{search_to_match}'
                )
        """
    elif search_type == "mailing_address":
        opa_account_numbers_sql = f"""
        SELECT opp.parcel_number FROM opa_properties_public opp
            WHERE (
                mailing_address_1 LIKE '{search_to_match}' 
                OR mailing_street LIKE '{search_to_match}'
            )
        """
    elif search_type == "location_by_owner":
        opa_account_numbers_sql = f"""
         SELECT opp2.parcel_number FROM opa_properties_public opp1, opa_properties_public opp2 
            WHERE opp1.location LIKE '{search_to_match}' AND (
                opp1.owner_1 = opp2.owner_1 OR opp1.owner_2 = opp2.owner_2
                OR opp1.owner_1 = opp2.owner_2 OR opp1.owner_2 = opp2.owner_1
            )
        """
    elif search_type == "location_by_mailing_address":
        opa_account_numbers_sql = f"""
         SELECT opp2.parcel_number FROM opa_properties_public opp1, opa_properties_public opp2 
            WHERE opp1.location LIKE '{search_to_match}' AND ( 
                opp1.mailing_street = opp2.mailing_street OR (
                    opp1.mailing_address_1 = opp2.mailing_address_1 
                    AND opp1.mailing_address_2 = opp2.mailing_address_2
                )
            )
        """
    else:
        raise SearchTypeNotImplementedError(search_type)

    return opa_account_numbers_sql



from scourgify import normalize_address_record
from passyunk.parser import PassyunkParser


def get_normalized_address(address):
    """Attempt to normalize an address string using two different cleaning libraries.

    Parameters
    ----------
    address: str
        A string representation of an address.

    Returns
    -------
    str
        A normalized string representation of an address.
    """
    p = PassyunkParser()
    # pass through philly-specific passyunk library
    address_pasyunk = p.parse(address)["components"]["output_address"]
    # attempt to further normalize
    n_address_dict = {
        k: v if v else "" for k, v in normalize_address_record(address).items()
    }
    # remove multiple spaces and combine
    return " ".join(
        (
            n_address_dict["address_line_1"]
            + n_address_dict["address_line_2"]
            + n_address_dict["city"]
            + n_address_dict["state"]
            + n_address_dict["postal_code"]
        )
        .strip()
        .split()
    )



def maybe_monkeypatch_response(
    monkeypatch, pytestconfig, carto_rows=None, response_override=None
):
    if not pytestconfig.getoption("make_request"):

        def _fake_get(*args, **kwargs):
            if response_override is not None:
                return MockResponse(response_override)
            elif "phl.carto.com" in args[0]:
                return MockCartoResponse(data=carto_rows)
            elif "https://api.phila.gov/tips/account/" in args[0]:
                return MockRealEstateTaxRevenueResponse()
            else:
                return MockResponse()

        monkeypatch.setattr("requests.get", _fake_get)


class MockResponse:
    def __init__(self, data=None, status_code=200):
        self.data = data
        self.status_code = status_code

    def json(self):
        return self.data


class MockRealEstateTaxRevenueResponse(MockResponse):
    def json(self):
        return {
            "data": {
                "accountNum": "01234",
                "property": {"abc": "def"},
                "lastPaymentPostedDate": "2020-01-01",
                "years": [{"def": "ghi", "year": 2020}],
            }
        }


class MockCartoResponse(MockResponse):
    def json(self):
        data = self.data if self.data is not None else [{"opa_account_num": "123"}]
        if self.data is None:
            data = [{"opa_account_num": "123"}]
        elif self.data:
            return {"rows": data, "fields": data[0]}
        else:
            return {"rows": [], "fields": {}}



def get_street_view_link(location):
    return f"https://cyclomedia.phila.gov/#/?address={location}"


def get_property_phila_gov_link(opa_account_number):
    return f"https://property-beta.phila.gov/#/?owner={opa_account_number}"


def get_atlas_link(opa_account_number):
    return f"https://atlas.phila.gov/{opa_account_number}"


def get_license_inspections_link(location):
    return f"https://li.phila.gov/property-history/search?address={location}"



import pandas as pd

from .tables import PhillyCartoTable


__all__ = (
    "Properties",
    "Permits",
    "Licenses",
    "Violations",
    "Condominiums",
    "Complaints",
    "Appeals",
    "RealEstateTaxDelinquencies",
    "RealEstateTransfers",
    "CaseInvestigations",
    "PropertiesPde",
)


class Properties(PhillyCartoTable):
    def __init__(self, title="Properties"):
        """
        This query always returns latitude and longitude in addition to any other
        columns requested.
        """
        super().__init__(
            cartodb_table_name="opa_properties_public",
            title=title,
            sql_alias="opa",  # other pieces of code rely on this to be 'opa'
            open_data_philly_table_url_name="opa-property-assessments",
            schema_application_id="550c60d00711ffe12e9efc64",
            schema_representation_id="55d624fdad35c7e854cb21a4",
        )
        self.default_columns = [
            "location",
            "unit",
            "owner_1",
            "owner_2",
            "mailing_care_of",
            "mailing_street",
            "mailing_address_1",
            "mailing_address_2",
            "mailing_city_state",
            "parcel_number",
            "building_code_description",
            "category_code_description",
            "homestead_exemption",
            "year_built",
            "year_built_estimate",
        ]

    def _get_sql_for_query_by_opa_account_numbers(
        self, opa_account_numbers, col_str, joined_col_str
    ):
        # no joining necessary since this is the primary table
        return f"""
        SELECT  {col_str}, ST_Y(the_geom) AS lat, ST_X(the_geom) AS lng
        FROM {self.cartodb_table_name} {self.sql_alias}
        WHERE parcel_number in ({opa_account_numbers})
        """

    def _get_sql_for_query_by_single_str_column(
        self,
        col_str,
        joined_col_str,
        search_column,
        search_to_match,
        limit_str,
    ):
        return f"""
        SELECT {col_str}, {joined_col_str}
        FROM {self.cartodb_table_name} {self.sql_alias}
        WHERE {self.sql_alias}.{search_column} LIKE '{search_to_match}' 
        {limit_str}
        """


class Licenses(PhillyCartoTable):
    def __init__(self, title="Licenses"):
        super().__init__(
            cartodb_table_name="business_licenses",
            title=title,
            open_data_philly_table_url_name="licenses-and-inspections-business-licenses",
            schema_application_id="5543ca6f5c4ae4cd66d3ff59",
            schema_representation_id="5e9a06fb98cc42001606f331",
        )
        self.default_columns = [
            "mostrecentissuedate",
            "licensestatus",
            "licensetype",
            "legalname",
            "business_name",
            "opa_owner",
            "numberofunits",
        ]
        self.dt_column = "mostrecentissuedate"


class Condominiums(PhillyCartoTable):
    def __init__(self, title="Condominiums"):
        super().__init__(
            cartodb_table_name="condominium",
            title=title,
        )
        self.default_columns = [
            "condo_name",
            "condounit",
            "orig_date",
            "inactdate",
            "status",
        ]

    def _get_sql_for_query_by_opa_account_numbers(
        self, opa_account_numbers, col_str, joined_col_str
    ):
        # involves an extra intermediary join through the rtt_summary
        return f"""
        SELECT {col_str}, {joined_col_str}
        FROM {self.cartodb_table_name} {self.sql_alias}
        LEFT JOIN rtt_summary rtt ON {self.sql_alias}.mapref=rtt.matched_regmap
        LEFT JOIN opa_properties_public opa ON rtt.opa_account_num=opa.parcel_number
        WHERE opa.parcel_number IN ({opa_account_numbers})
            AND rtt.unit_num = {self.sql_alias}.condounit
        """


class Complaints(PhillyCartoTable):
    def __init__(self, title="Complaints"):
        super().__init__(
            cartodb_table_name="complaints",
            title=title,
            open_data_philly_table_url_name="licenses-and-inspections-service-requests",
            schema_application_id="5543ca6d5c4ae4cd66d3ff52",
            schema_representation_id="5e5d50e0fbc9650019b56025",
        )
        self.default_columns = [
            "complaintdate",
            "complaintnumber",
            "complaintcodename",
            "complaintdate",
            "complaintstatus",
            "casestatus",
            "initialinvestigation_date",
            "complaintresolution_date",
        ]
        self.dt_column = "complaintdate"


class Violations(PhillyCartoTable):
    def __init__(self, title="Violations"):
        super().__init__(
            cartodb_table_name="violations",
            title=title,
            open_data_philly_table_url_name="licenses-and-inspections-violations",
            schema_application_id="5543ca7a5c4ae4cd66d3ff86",
            schema_representation_id="5e99bab227c80700158695b0",
        )
        self.default_columns = [
            "violationdate",
            "caseprioritydesc",
            "violationcode",
            "violationcodetitle",
        ]
        self.dt_column = "violationdate"


class Permits(PhillyCartoTable):
    def __init__(self, title="Permits"):
        super().__init__(
            cartodb_table_name="permits",
            title=title,
            open_data_philly_table_url_name="licenses-and-inspections-building-permits",
            schema_application_id="5543868920583086178c4f8f",
            schema_representation_id="5e9a01ac801624001585ca11",
        )
        self.default_columns = [
            "permitissuedate",
            "permitnumber",
            "permitdescription",
            "typeofwork",
            "approvedscopeofwork",
            "status",
            "applicanttype",
            "contractorname",
            "contractoraddress1",
            "contractoraddress2",
            "contractorcity",
            "contractorstate",
            "contractorzip",
            "mostrecentinsp",
        ]
        self.dt_column = "permitissuedate"


class Appeals(PhillyCartoTable):
    def __init__(self, title="Appeals"):
        super().__init__(
            cartodb_table_name="appeals",
            title=title,
            open_data_philly_table_url_name="license-and-inspections-appeals",
            schema_application_id="5543864d20583086178c4e9c",
            schema_representation_id="5e9751361ed3930016d62645",
        )
        self.default_columns = [
            "createddate",
            "primaryappellant",
            "appellanttype",
            "appealnumber",
            "acceleratedappeal",
            "completeddate",
            "relatedpermit",
            "relatedcasefile",
            "appealstatus",
            "appealtype",
            "agendadescription",
            "applicationtype",
            "meetingnumber",
            "appealgrounds",
            "scheduleddate",
            "decision",
            "meetingresult",
            "proviso",
            "decisiondate",
        ]

        self.dt_column = "createddate"


class RealEstateTaxDelinquencies(PhillyCartoTable):
    def __init__(self, title="Real Estate Tax Delinquencies"):
        super().__init__(
            cartodb_table_name="real_estate_tax_delinquencies",
            title=title,
            open_data_philly_table_url_name="property-tax-delinquencies",
            schema_application_id="57d9643afab162fe2708224e",
            schema_representation_id="57d9643cfab162fe27082252",
        )
        self.default_columns = [
            "most_recent_year_owed",
            "street_address",
            "total_due",
            "num_years_owed",
            "most_recent_payment_date",
            "coll_agency_num_years",
            "years_in_bankruptcy",
            "building_category",
        ]

        self.dt_column = "most_recent_year_owed"

    def _get_sql_for_query_by_opa_account_numbers(
        self, opa_account_numbers, col_str, joined_col_str
    ):
        # involves casting opa_number to text
        return f"""
        SELECT {col_str}, {joined_col_str}
        FROM {self.cartodb_table_name} {self.sql_alias}
        LEFT JOIN opa_properties_public opa 
            ON cast({self.sql_alias}.opa_number as text)=opa.parcel_number
        WHERE opa.parcel_number in ({opa_account_numbers})
        """


class RealEstateTransfers(PhillyCartoTable):
    def __init__(self, title="Real Estate Transfers"):
        super().__init__(
            cartodb_table_name="rtt_summary",
            title=title,
            open_data_philly_table_url_name="real-estate-transfers",
            schema_application_id="5a04b8d39202605970a7457d",
            schema_representation_id="5a04b8d39202605970a74581",
        )
        self.default_columns = [
            "receipt_date",
            "street_address",
            "grantors",
            "grantees",
            "total_consideration",
            "condo_name",
            "unit_num",
            "receipt_date",
            "recording_date",
            "document_id",
            "document_type",
        ]

        self.dt_column = "receiptdate"

    def infer_property_ownership(self, opa_account_number, recording_date=None):
        """Infer property ownership at a specific point in time

        Based on the owner of the deed at the time, provides the owner name (via the
        grantee field) for a given date at a given property.

        This is done using the following rules:

        1) Search for all of the DEED transactions that took place at a property
        2) reconstruct the address from the address components
        3) compare this reconstructed address with the provided `street_address` col
            3a) This is because we have found DEEDs that were associated with
                an opa_account_num but didn't turn out to be actually for that
                property.
        4) Look at the recording_date provided (or use datetime.now() if None provided)
        5) Find the most recent DEED grantee before the given recording_date
            5a) If None is found (the rtt_summary data doesn't start until 1999),
                then use the earliest grantor in the DEED data provided.

        This function introduces the concept of 'discrepencies' which may be
        expanded elsewhere in this library. These are assumptions/estimates/deviations
        that were necessary to return a result, but may not be as reliable as otherwise
        standard results. 3a and 5a above are examples of discrepencies.

        Parameters
        ----------
        opa_account_num: str
            The opa_account_num for the property.
        recording_date: str
            The date to look for property ownership.
        """

        recording_date = (
            recording_date if recording_date else datetime.now().isoformat()
        )
        list_result = self.list(
            columns=[
                "opa_account_num",
                "recording_date",
                "grantors",
                "grantees",
                "address_low",
                "address_low_suffix",
                "address_high",
                "address_low_frac",
                "street_predir",
                "street_name",
                "street_suffix",
                "street_address",
            ],
            where_sql=f"""
            (
                document_type='DEED' OR
                document_type='DEED SHERIFF' OR
                document_type='DEED OF CONDEMNATION' OR
                document_type='DEED LAND BANK'
            ) AND opa_account_num ='{opa_account_number}'
            """,
            order_by_columns=["recording_date"],
        )
        df = list_result.to_dataframe()
        if df.empty:
            df_location = (
                Properties()
                .query_by_opa_account_numbers([opa_account_number])
                .to_dataframe()
            )
            if not df_location.empty:
                location_address = df_location.iloc[0]["location"]

        """
        (((ADDRESS_LOW >= 5427 AND ADDRESS_LOW <= 5427)
        OR (ADDRESS_LOW >= 5400 AND ADDRESS_LOW <= 5427 AND ADDRESS_HIGH >= 27 ))
        AND STREET_NAME = 'WAYNE' AND STREET_SUFFIX = 'AVE' AND (MOD(ADDRESS_LOW,2) = MOD( 5427,2)))
        """
        return self.infer_property_ownership_from_df(
            df, opa_account_number, recording_date
        )

    def infer_property_ownership_from_df(self, df, opa_account_number, recording_date):
        def _compose_address(x):
            def _strfy(col, prefix=""):
                return prefix + str(x.loc[col]) if x.loc[col] else ""

            address_low = _strfy("address_low")
            address_low_suffix = _strfy("address_low_suffix")
            address_high = _strfy("address_high", prefix="-")
            address_low_frac = _strfy("address_low_frac")
            street_predir = _strfy("street_predir")
            street_name = _strfy("street_name")
            street_suffix = _strfy("street_suffix")
            full_address = (
                f"{address_low}{address_low_suffix}{address_high} {address_low_frac}"
                f" {street_predir} {street_name} {street_suffix}"
            )
            return " ".join(full_address.split())

        discrepencies = []
        if not df.empty:
            df["composed_address"] = df.apply(_compose_address, axis=1)
            if not df[df["composed_address"] != df["street_address"]].empty:
                # If it finds a non-matching address:
                data = df[df["composed_address"] != df["street_address"]][
                    [
                        "composed_address",
                        "street_address",
                        "opa_account_num",
                        "recording_date",
                    ]
                ].to_dict("records")
                discrepencies.append(
                    {
                        "data": data,
                        "description": "There was an address associated with a DEED "
                        "that didn't seem to match the property.",
                    }
                )
            df_match = df[df["composed_address"] == df["street_address"]]
        if df.empty or df_match.empty:
            discrepencies.append(
                {
                    "data": {
                        "opa_account_num": opa_account_number,
                        "recording_date": recording_date,
                    },
                    "description": "There is no DEED information available for this property.",
                }
            )
            return {
                "owner": None,
                "metadata": None,
                "discrepencies": discrepencies,
            }
        else:
            df_before = df_match[df_match["recording_date"] < recording_date]
            if df_before.empty:
                discrepencies.append(
                    {
                        "description": (
                            "The DEED for this property during this time was transferred"
                            " earlier than this dataset has acccess (before 1999)."
                        )
                    }
                )
                df_out = df_match.copy().iloc[0]
                df_out["owner"] = df_out["grantors"]
            else:
                df_out = df_before.copy().iloc[-1]
                df_out["owner"] = df_out["grantees"]
        output_dict = df_out[
            [
                "owner",
                "grantors",
                "grantees",
                "recording_date",
                "opa_account_num",
                "composed_address",
                "street_address",
            ]
        ].to_dict()
        return {
            "owner": output_dict["owner"],
            "metadata": output_dict,
            "discrepencies": discrepencies,
        }


class CaseInvestigations(PhillyCartoTable):
    def __init__(self, title="Case Investigations"):
        super().__init__(
            cartodb_table_name="case_investigations",
            title=title,
            open_data_philly_table_url_name="licenses-and-inspections-inspections",
            schema_application_id="5543ca785c4ae4cd66d3ff80",
            schema_representation_id="5e986970b2c39b001522fb9d",
        )
        self.default_columns = [
            "investigationcompleted",
            "casenumber",
            "investigationprocessid",
            "casetype",
            "caseresponsibility",
            "casepriority",
            "investigationtype",
            "investigationstatus",
        ]

        self.dt_column = "investigationcompleted"


class PropertiesPde(PhillyCartoTable):
    def __init__(self, title="Properties (Cleaned)"):
        """
        This query always returns latitude and longitude in addition to any other
        columns requested.
        """
        super().__init__(
            cartodb_table_name="opa_properties_public_pde",
            title=title,
            sql_alias="opa",  # other pieces of code rely on this to be 'opa'
        )
        self.default_columns = [
            "address_std",
            "location",
            "unit",
            "owner_1",
            "owner_2",
            "mailing_care_of",
            "mailing_street",
            "mailing_address_1",
            "mailing_address_2",
            "mailing_city_state",
            "parcel_number",
            "political_district",
            "police_district",
            "elementary_school",
            "middle_school",
            "high_school",
            "parcel_number",
        ]

    def _get_sql_for_query_by_opa_account_numbers(
        self, opa_account_numbers, col_str, joined_col_str
    ):
        # no joining necessary
        return f"""
        SELECT  {col_str}, ST_Y(the_geom) AS lat, ST_X(the_geom) AS lng
        FROM {self.cartodb_table_name} {self.sql_alias}
        WHERE parcel_number in ({opa_account_numbers})
        """



CITY_OWNED_PROPERTY_FIELDS = {
    "owner_1": [
        "CITY OF PHILA",
        "PHILA HOUSING AUTHORITY",
        "CITY OF PHILADELPHIA",
        "PHILADELPHIA HOUSING",
        "REDEVELOPMENT AUTHORITY",
        "REDEVELOPMENT AUTHORITY O",
        "SCHOOL DISTRICT OF PHILA",
        "PHILADELPHIA HOUSING AUTH",
        "PHILA HOUSING AUTHORITY",
        "PHILADELPHIA LAND BANK",
        "PHILADELPHIA HOUSING DEVE",
        "CITY OC PHILA",
        "CITY HALL ANNEX",
    ],
    "mailing_street": [
        "12 S 23RD ST",
        "1401 JOHN F KENNEDY BLVD",
        "790 CITY HALL",
        "CITY HALL ANNEX",
        "CITY HALL RM 784",
        "CITY HALL",
        "723 CITY HALL ANNEX",
        "3100 PENROSE FERRY RD",
    ],
}


def label_city_owned_properties(df, remove=False, city_owned_label="is_city_owned"):
    """Checks to see if the property is owned by the city.

    This requires the input dataframe to have columns specified in
    CITY_OWNED_PROPERTY_FIELDS

    Parameters
    ----------
    df: pd.DataFrame
        A dataframe of properties that contains data allowing it to be labelled as
        city-owned or not.
    remove: bool
        Whether to remove properties identified as city-owned from the returned dataframe.
    city_owned_label: str
        The column name to include the boolean of whether each property is city-owned.

    Returns
    -------
    pd.DataFrame
        Returns a dataframe with a `is_city_owned

    """
    city_owned_required_cols = CITY_OWNED_PROPERTY_FIELDS.keys()
    for column, values in CITY_OWNED_PROPERTY_FIELDS.items():
        if column not in df.columns:
            raise ValueError(
                f"{column} not in results. You must "
                f"request {city_owned_required_cols} columns "
                "in order to check for city-owned properties."
            )
        city_owned_query = f"{column}.str.strip() not in {tuple(values)}"
        if remove:
            df = df.query(city_owned_query)
        else:
            df[city_owned_label] = False
            df.loc[df.query(city_owned_query)] = True
    return df



class SearchTypeNotImplementedError(Exception):
    """Exception raised for errors where search type provided isn't implemented."""

    def __init__(self, search_type):
        self.message = f"""
            Invalid search_type: '{search_type}' provided. Must be one of: ('owner',
            'mailing_address', 'location_by_mailing_address', 'location_by_owner',
            'parcel_number')
            """.replace(
            "\n", " "
        )
        super().__init__(self.message)


class SearchMethodNotImplementedError(Exception):
    """Exception raised for errors where search method provided isn't implemented."""

    def __init__(self, search_method):
        self.message = (
            f"Invalid search_method: '{search_method}' provided. Must be one of: "
            "('contains', 'starts_with', 'ends_with')"
        )
        super().__init__(self.message)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
NAME = "phillydb"
about = {}

with open(os.path.join(here, NAME, "__version__.py")) as f:
    exec(f.read(), about)

setup(
    name=NAME,
    version=about["__version__"],
    description=about["__description__"],
    entry_points={"console_scripts": [f"{NAME} = {NAME}:cli"]},
    packages=find_packages(),
    install_requires=[
        "pandas",
        "requests",
        "passyunk",
        "usaddress-scourgify",
        "fuzzywuzzy",
    ],
)
